# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "ProjectError",
    "ProjectNotFoundError",
]

import functools
import json
import shutil
import tempfile
import time
import traceback
import warnings
import zipfile
from dataclasses import asdict
from pathlib import Path
from urllib.request import urlretrieve
from uuid import uuid4

import jsonschema
import numpy as np
import scipy.sparse as sp
from filelock import FileLock

from asreview.data.loader import _from_file
from asreview.data.loader import _get_reader
from asreview.data.utils import identify_record_groups
from asreview.database.database import Database
from asreview.datasets import DatasetManager
from asreview.learner import ActiveLearningCycle
from asreview.learner import ActiveLearningCycleData
from asreview.models import get_ai_config
from asreview.project.exceptions import ProjectError
from asreview.project.exceptions import ProjectNotFoundError
from asreview.project.migration import detect_version
from asreview.project.migration import migrate_project
from asreview.project.schema import SCHEMA
from asreview.utils import _get_filename_from_url
from asreview.utils import _is_url

try:
    from asreview._version import __version__
except ImportError:
    __version__ = "0.0.0"


def is_project(project_dir):
    """
    Check if the given path is a valid ASReview project.

    Parameters
    ----------
    project_dir : str | Path
        The path to the project directory.

    Returns
    -------
    bool
        True if the path is a valid ASReview project, False otherwise.
    """
    project_dir = Path(project_dir)
    if not project_dir.exists():
        return False
    if not project_dir.is_dir():
        return False
    project_config_fp = Path(project_dir, Project.PATH_CONFIG)
    if not project_config_fp.exists():
        return False
    with open(project_config_fp) as f:
        project_config = json.load(f)
    if detect_version(project_config) != Project.VERSION:
        return False
    return True


class Project:
    """Project class for ASReview project files.

    This class represents the complete data file for a review project. This data is
    contained in a single directory with the following files and subdirectories:
    - `project.json`: A JSON file containing the configuration and metadata of the
    project. It's structure is described in `schema.py`
    - `data/`: A directory containing the input data file exactly as provided by the
    user. When exporting, this input data is merged with the results of the review to
    get the export file.
    - `feature_matrices/`: A directory containing all the feature matrices that are
    generated during the review.
    - `results.db`: An SQLite database containing all data generated by ASReview: the
    data parsed from the input file, the labeled records, the last model ranking etc.
    See `asreview/data` for information on the parsing of the input. See
    `asreview/state` for information on the model and the labeling decisions.
    """

    VERSION = 3
    MODE_SIMULATE = "simulate"
    PATH_CONFIG = "project.json"
    PATH_CONFIG_LOCK = "project.json.lock"
    PATH_FEATURE_MATRICES = "feature_matrices"
    PATH_DATA_DIR = "data"
    PATH_DB = "results.db"
    PATH_ERROR = "error.json"

    def __init__(self, project_path, project_id=None):
        self.project_path = Path(project_path)
        self.project_id = project_id
        self.data_dir = Path(self.project_path, self.PATH_DATA_DIR)
        self.db_path = Path(self.project_path, self.PATH_DB)
        self.error_path = Path(self.project_path, self.PATH_ERROR)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """Close the project and release all resources.

        Closes the database connection if it was opened. Safe to call multiple
        times.
        """
        if "db" in self.__dict__:
            self.db.close()

    @functools.cached_property
    def db(self):
        return Database(self.db_path)

    @property
    def input_data_fp(self):
        datasets = self.config.get("datasets")
        if not datasets:
            return
        file_name = datasets[0]["name"]
        return self.data_dir / file_name

    @classmethod
    def create(
        cls,
        project_path,
        project_id=None,
        project_mode="oracle",
        project_name=None,
        project_tags=None,
    ):
        """Initialize the necessary files specific to the web app."""

        project_path = Path(project_path)

        if project_path.exists():
            raise ValueError("Project path is not empty.")

        if project_id is None:
            project_id = project_path.stem

        if project_name is None:
            project_name = project_path.stem

        if project_path.is_dir():
            raise IsADirectoryError(f"Project folder {project_path} already exists.")

        try:
            project_path.mkdir(parents=True, exist_ok=True)
            Path(project_path, "data").mkdir(exist_ok=True)
            Path(project_path, cls.PATH_FEATURE_MATRICES).mkdir(exist_ok=True)
            with Database(Path(project_path, cls.PATH_DB)) as database:
                database.create_tables()

            config = {
                "version": __version__,
                "id": project_id,
                "mode": project_mode,
                "name": project_name,
                "created_at_unix": int(time.time()),
                "feature_matrices": [],
                "tags": project_tags,
                "project_file_version": cls.VERSION,
            }

            jsonschema.validate(instance=config, schema=SCHEMA)

            project_fp = Path(project_path, cls.PATH_CONFIG)
            project_fp_lock = Path(project_path, cls.PATH_CONFIG_LOCK)
            lock = FileLock(project_fp_lock, timeout=3)

            with lock:
                with open(project_fp, "w") as f:
                    json.dump(config, f)

        except Exception as err:
            shutil.rmtree(project_path)
            raise err

        return cls(project_path, project_id=project_id)

    @property
    def config(self):
        try:
            return self._config
        except AttributeError:
            project_fp = Path(self.project_path, self.PATH_CONFIG)
            project_fp_lock = Path(self.project_path, self.PATH_CONFIG_LOCK)
            lock = FileLock(project_fp_lock, timeout=3)

            if not project_fp.exists():
                raise ProjectNotFoundError(f"Project '{self.project_path}' not found")

            with lock:
                # read the file with project info
                with open(project_fp) as fp:
                    config = json.load(fp)
                    self._config = config

                    return config

    @config.setter
    def config(self, config):
        project_fp = Path(self.project_path, self.PATH_CONFIG)
        project_fp_lock = Path(self.project_path, self.PATH_CONFIG_LOCK)
        lock = FileLock(project_fp_lock, timeout=3)

        with lock:
            with open(project_fp, "w") as f:
                json.dump(config, f)

        self._config = config

    def update_config(self, **kwargs):
        """Update project info"""

        config = self.config
        config.update(kwargs.copy())

        jsonschema.validate(instance=config, schema=SCHEMA)

        self.config = config
        return config

    def add_dataset(self, fp, dataset_id=None, file_writer=None):
        """Add a dataset to the project file.

        Parameters
        ----------
        fp: str, Path
            Filepath to the dataset. It will be copied to the correct location in the
            project file.
        """
        if dataset_id is None:
            dataset_id = uuid4().hex

        self.data_dir.mkdir(exist_ok=True)

        if file_writer is not None:
            save_fp = self.data_dir / fp
            file_writer(save_fp)
        elif _is_url(fp):
            filename = _get_filename_from_url(fp)
            save_fp = self.data_dir / filename
            urlretrieve(fp, save_fp)
        elif Path(fp).exists():
            save_fp = self.data_dir / Path(fp).name
            shutil.copy(fp, save_fp)
        else:
            dataset = DatasetManager().find(fp)
            if dataset is None:
                raise ValueError(
                    "fp should be existing file, or URL or dataset, but does not"
                    f" exist: {fp}"
                )
            save_fp = self.data_dir / dataset.filename
            dataset.to_file(save_fp)
        file_name = save_fp.name

        records = _from_file(save_fp, dataset_id=dataset_id)

        # Internals of the records are leaking out here. We are checking for a specific
        # field and a specific value. If the presence of the field `included` is
        # necessary in the input data, we should move it from `Record` to the `Base`
        # class, so that all record implementations have it.
        if self.config["mode"] == self.MODE_SIMULATE and (
            all([r.included is None for r in records])
        ):
            raise ValueError(
                "Dataset for simulation mode must have labels for all records - "
                "got dataset without any labels"
            )

        if self.config["mode"] == self.MODE_SIMULATE and (
            any([r.included is None for r in records])
        ):
            raise ValueError(
                "Dataset for simulation mode must be fully labeled - "
                "got records with missing labels"
            )

        self.db.input.add_records(records=records)
        groups = identify_record_groups(records)
        self.db.input.set_groups(groups)

        # This config update assumes that the project only has one dataset.
        self.update_config(
            name=file_name.rsplit(".", 1)[0],
            datasets=[{"id": dataset_id, "name": file_name}],
        )

    def remove_dataset(self):
        """Remove dataset from project."""
        raise NotImplementedError("Removing datasets is not implemented yet")

    def read_input_data(self, *args, **kwargs):
        reader = self.get_input_data_reader()
        return reader.read_data(self.input_data_fp, *args, **kwargs)

    def get_input_data_reader(self):
        return _get_reader(self.input_data_fp)

    @property
    def feature_matrices(self):
        try:
            return self.config["feature_matrices"]
        except Exception:
            return []

    def add_feature_matrix(self, feature_matrix, name):
        """Add feature matrix to project file.

        Parameters
        ----------
        feature_matrix: numpy.ndarray, scipy.sparse.csr.csr_matrix
            The feature matrix to add to the project file.
        name: str
            Name of the feature extractor.
        """
        file_name = f"{name}_feature_matrix"
        file_path = Path(self.project_path, self.PATH_FEATURE_MATRICES, file_name)

        if sp.issparse(feature_matrix):
            sp.save_npz(str(file_path), feature_matrix)
            file_name += ".npz"
        elif isinstance(feature_matrix, np.ndarray):
            np.save(file_path, feature_matrix)
            file_name += ".npy"
        elif isinstance(feature_matrix, list):
            np.save(file_path, np.array(feature_matrix))
            file_name += ".npy"
        else:
            raise ValueError("Unsupported feature matrix type")

        # Add the feature matrix to the project config.
        config = self.config

        feature_matrix_config = {
            "id": name,
            "filename": file_name,
        }

        # Add container for feature matrices.
        if "feature_matrices" not in config:
            config["feature_matrices"] = []

        config["feature_matrices"].append(feature_matrix_config)

        self.config = config

    def get_feature_matrix(self, name):
        """Get the feature matrix from the project file.

        Parameters
        ----------
        name : str
            Name of the feature extractor for which to get the cached matrix.

        Returns
        -------
        numpy.ndarray, scipy.sparse:
            (Sparse) feature matrix.
        """
        feature_matrix_config = [
            x for x in self.config["feature_matrices"] if x["id"] == name
        ]

        if len(feature_matrix_config) == 0:
            raise ValueError("Feature matrix not found")

        file_path = Path(
            self.project_path,
            self.PATH_FEATURE_MATRICES,
            feature_matrix_config[0]["filename"],
        )

        if file_path.suffix == ".npz":
            return sp.load_npz(str(file_path))
        elif file_path.suffix == ".npy":
            return np.load(file_path, allow_pickle=False)
        else:
            raise ValueError("Unsupported file extension")

    @property
    def review(self):
        return self.config.get("review")

    def add_review(
        self,
        cycle=None,
        reviewer=None,
        status="setup",
    ):
        """Add new review metadata.

        Parameters
        ----------
        cycle:
            An active learning cycle object to add to the review. This object is used
            to store the configuration of the active learning cycle to file.
        reviewer: object
            A reviewer object with to_sql() method.
        status: str
            The status of the review. One of 'setup', 'running',
            'finished'.

        """
        if self.review is not None:
            raise ValueError("Review already exists.")

        self.update_review(model=cycle, status=status)

        if reviewer is not None:
            reviewer.to_sql(self.db_path)

        return self.config

    def update_review(self, status=None, model_name=None, model=None):
        """Update review metadata."""
        review_config = self.config.get("review", {"status": "setup", "model": {}})
        if status is not None:
            review_config["status"] = status
        if model is not None:
            if not isinstance(
                model, (ActiveLearningCycle, ActiveLearningCycleData, dict)
            ):
                raise ValueError(
                    "model should be of type 'dict', 'ActiveLearningCycle' or "
                    "'ActiveLearningCycleData'"
                )
            if isinstance(model, ActiveLearningCycle):
                model = model.to_meta()
            if isinstance(model, ActiveLearningCycleData):
                model = asdict(model)
            review_config["model"]["current_value"] = model
        if model_name is not None:
            review_config["model"]["name"] = model_name
        self.update_config(review=review_config)

    def get_model_config(self):
        """Get the current model configuration of the review.

        Returns
        -------
        dict | None
            Dictionary containing the model configuration. Returns None if there is no
            review yet in the project.
        """
        return self.config.get("review", {}).get("model", {}).get("current_value")

    def export(self, export_fp):
        if Path(export_fp).suffix != ".asreview":
            raise ValueError("Export file should have .asreview extension.")

        if Path(export_fp) == Path(self.project_path):
            raise ValueError("export_fp should not be identical to project path.")

        export_fp_tmp = Path(export_fp).with_suffix(".asreview.zip")

        # copy the source tree, but ignore pickle files
        shutil.copytree(
            self.project_path,
            export_fp_tmp,
            ignore=shutil.ignore_patterns("tmp", "*.lock"),
        )

        # create the archive
        shutil.make_archive(export_fp_tmp, "zip", root_dir=export_fp_tmp)

        # remove the unzipped folder and move zip
        shutil.rmtree(export_fp_tmp)
        shutil.move(f"{export_fp_tmp}.zip", export_fp)

    @classmethod
    def load(
        cls,
        asreview_file,
        project_path,
        safe_import=False,
        reset_model_if_not_found=False,
    ):
        with tempfile.TemporaryDirectory() as tmpdir:
            try:
                # Unzip the project file
                with zipfile.ZipFile(asreview_file, "r") as zip_obj:
                    zip_filenames = zip_obj.namelist()

                    # raise error if no ASReview project file
                    if cls.PATH_CONFIG not in zip_filenames:
                        raise ValueError("Project file is not valid project.")

                    # extract all files to folder
                    for f in zip_filenames:
                        if not (f.endswith(".pickle") or f.endswith(".lock")):
                            zip_obj.extract(f, path=tmpdir)

            except zipfile.BadZipFile:
                raise ValueError("File is not an ASReview file.")

            with open(Path(tmpdir, cls.PATH_CONFIG)) as f:
                project_config = json.load(f)

            # if migration is needed, do it here
            current_version = detect_version(project_config)
            if current_version < cls.VERSION:
                migrate_project(tmpdir, current_version, cls.VERSION)
                # Migration may have updated project_config, so we reload it.
                with open(Path(tmpdir, cls.PATH_CONFIG)) as f:
                    project_config = json.load(f)

            if reset_model_if_not_found:
                try:
                    model_config = (
                        project_config.get("review", {})
                        .get("model", {})
                        .get("current_value")
                    )
                    ActiveLearningCycle.from_meta(
                        ActiveLearningCycleData(**model_config)
                    )
                except ValueError as err:
                    warnings.warn(str(err))
                    model = get_ai_config()
                    project_config["review"]["model"] = {
                        "name": model["name"],
                        "current_value": asdict(model["value"]),
                    }
                    with open(Path(tmpdir, cls.PATH_CONFIG), "w") as f:
                        json.dump(project_config, f)

            if safe_import:
                # assign a new id to the project.
                project_config["id"] = uuid4().hex
                with open(Path(tmpdir, cls.PATH_CONFIG), "w") as f:
                    json.dump(project_config, f)

            shutil.copytree(tmpdir, Path(project_path, project_config["id"]))
        return cls(Path(project_path, project_config["id"]))

    def get_review_error(self):
        if self.error_path.exists():
            with open(self.error_path, "r") as f:
                return json.load(f)
        else:
            raise ValueError("No error found.")

    def set_review_error(self, err):
        err_type = type(err).__name__

        with open(self.error_path, "w") as f:
            json.dump(
                {
                    "message": f"{err_type}: {err}",
                    "type": f"{err_type}",
                    "time": int(time.time()),
                    "traceback": traceback.format_exc(),
                },
                f,
            )

    def remove_review_error(self):
        self.error_path.unlink(missing_ok=True)



from typing import Optional

import pandas as pd
from sqlalchemy import CheckConstraint, Computed, Index
from sqlalchemy import ForeignKey
from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import DeclarativeBase, declared_attr
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import validates
from sqlalchemy.types import JSON
from sqlalchemy.types import Integer
from sqlalchemy.types import String

from asreview.data.utils import convert_value_to_int


class Base(DeclarativeBase, MappedAsDataclass):
    """Base class for records.

    Anyone implementing a custom record type for ASReview should inherit this class.
    This class inherits `DeclarativeBase` which indicates that we are using the
    declarative way of defining models in SQLAlchemy ORM. See
    'https://docs.sqlalchemy.org/en/20/orm/' for more information.

    This class also inherits `MappedAsDataclass`, which means that any class that
    inherits `Base` will automatically be a Python dataclass (i.e. it has
    auto-implementations of `__init__`, `__eq__`, `__repr__`, etc.), and we can call
    `dataclasses.asdict` to turn in instance into a dictionary.

    There are a number of attributes you can customize to make `DataStore` understand
    how to store a record in the database and how to read record data into pandas
    objects. See the comments in the source code.
    """

    # `DataStore` needs to read data from the database and put it in a Pandas dataFrame
    # or series. It needs to know into which datatype each column should be converted.
    # This mapping specifies the conversion from SQLAlchemy datatype to Pandas datatype.
    __sqlalchemy_to_pandas_dtype_mapping__ = {
        Integer: "Int64",
        String: "object",
        JSON: "object",
    }

    # When storing a record, the database needs to know what datatype each field should
    # have. Some types have default values, like `int` for example, but other do not.
    # The following mapping specifies the conversion from Python datatypes to SQLAlchemy
    # datatypes. See https://docs.sqlalchemy.org/en/20/orm/declarative_tables.html#customizing-the-type-map.
    type_annotation_map = {list: JSON}

    # The primary identifier of the record. We always want this to correspond to the
    # row_number of the record in the feature matrix.
    # Ideally this field should be called 'id' instead of 'record_id', but that requires
    # refactoring in the frontend as well, so I'll skip that for now.
    record_id: Mapped[int] = mapped_column(init=False, primary_key=True)

    # The following column depends on the table name of the concrete class using this
    # base class, so it needs to be a declared attribute. The `group_id` column and the
    # index on `group_id` also need to be declared attributes because of that.
    @declared_attr
    def duplicate_of(cls) -> Mapped[Optional[int]]:
        return mapped_column(
            Integer,
            ForeignKey(f"{cls.__tablename__}.record_id", ondelete="SET NULL"),
            init=False,
            default=None,
        )

    @declared_attr
    def group_id(cls) -> Mapped[int]:
        return mapped_column(
            Computed("COALESCE(duplicate_of, record_id)", persisted=False),
            init=False,
            default=None,
        )

    # Concrete classes using this base class can implement extra table args using the
    # `__extra_table_args__` attribute.
    @declared_attr
    def __table_args__(cls):
        extra = getattr(cls, "__extra_table_args__", ())
        return (
            Index(f"idx_{cls.__tablename__}_group_id", "group_id"),
            *extra,
        )

    @classmethod
    def get_columns(cls):
        """Get the list of database column names.

        Returns
        -------
        list[str]
        """
        return [column.name for column in cls.__mapper__.columns]

    @classmethod
    def get_pandas_dtype_mapping(cls):
        """Get the mapping from record column name to pandas data type.

        By default it uses the mapping from `__sqlalchemy_to_pandas_dtype_mapping__`,
        but you can overwrite this method if you need different behavior.

        Returns
        -------
        dict[str, str]
            Dictionary whose keys are record column names and the values are the
            corresponding pandas data type that the column should have when reading it
            into a pandas object.
        """
        return {
            column.name: cls.__sqlalchemy_to_pandas_dtype_mapping__[
                column.type.__class__
            ]
            for column in cls.__mapper__.columns
        }


class Record(Base):
    __tablename__ = "record"
    __extra_table_args__ = (
        UniqueConstraint("dataset_row", "dataset_id"),
        CheckConstraint(
            "duplicate_of IS NULL OR duplicate_of != record_id",
            name="no_self_duplicate",
        ),
    )
    # We use dataset_row to locate the record in the original input file of the user.
    # For now I call this 'row', meaning that we will look in the input file by row
    # number. We might want to change this to locate the record by an external
    # identifier provided by the user?
    # Also, maybe we want these two to be on `Base` and not on `Record`? Then anyone
    # implementing a custom record is forced to have them.
    dataset_row: Mapped[int]
    dataset_id: Mapped[str]

    title: Mapped[str] = mapped_column(default="")
    abstract: Mapped[str] = mapped_column(default="")
    # authors and keywords could also be in their own separate table.
    authors: Mapped[list] = mapped_column(default_factory=list)
    keywords: Mapped[list] = mapped_column(default_factory=list)
    year: Mapped[Optional[int]] = mapped_column(default=None)
    doi: Mapped[Optional[str]] = mapped_column(default=None)
    url: Mapped[Optional[str]] = mapped_column(default=None)
    included: Mapped[Optional[int]] = mapped_column(default=None)

    @validates("authors", "keywords")
    def validate_list_of_string(self, key, value):
        if value is None:
            return []
        if not isinstance(value, list):
            raise ValueError(f"'{key}' should be a list or None, but is: {value}")
        if not all(isinstance(item, str) for item in value):
            raise ValueError(f"'{key}' should be a list of strings")
        return value

    @validates("title", "abstract", "doi", "url")
    def validate_string(self, key, value):
        if value == "":
            return None
        if value is not None and not isinstance(value, str):
            raise ValueError(f"'{key}' should be a string or None, but is: {value}")
        return value

    @validates("included")
    def validate_included(self, key, included):
        if pd.isna(included):
            included = None

        if not (included is None or included in {0, 1}):
            raise ValueError(
                f"included should be one of 0, 1, or None. Not '{included}'"
            )
        return included

    @validates("year")
    def validate_optional_integer(self, key, value):
        return convert_value_to_int(value)



import json
from dataclasses import asdict
from pathlib import Path

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

import asreview as asr
from asreview.simulation.cli import _cli_simulate


def test_dataset_not_found(tmp_project):
    argv = f"does_not.exist -o {tmp_project}".split()
    with pytest.raises(ValueError):
        _cli_simulate(argv)


def test_no_output(demo_data_path):
    _cli_simulate([str(demo_data_path)])


def test_simulate_review_finished(tmpdir, demo_data_path, tmp_project):
    _cli_simulate(f"{demo_data_path} -o {tmp_project}".split())

    Path(tmpdir, "test").mkdir(parents=True)
    project = asr.Project.load(tmp_project, Path(tmpdir, "test"))
    assert project.config["review"]["status"] == "finished"


def test_prior_idx(tmp_project, demo_data_path, tmpdir):
    argv = f"{demo_data_path} -o {tmp_project} --prior-idx 0 9".split()
    _cli_simulate(argv)

    with asr.Project.load(tmp_project, tmpdir).db as db:
        results_table = db.get_results_table()

    assert results_table["record_id"].head(2).to_list() == [0, 9]
    assert results_table["querier"][:1].isnull().all()
    assert results_table["querier"][2:].notnull().all()


def test_prior_record_ids(tmp_project, demo_data_path, tmpdir):
    argv = f"{demo_data_path} -o {tmp_project} --prior-record-id 0 9".split()
    _cli_simulate(argv)

    with asr.Project.load(tmp_project, tmpdir).db as db:
        results_table = db.get_results_table()

    assert results_table["record_id"].head(2).to_list() == [0, 9]
    assert results_table["querier"][:1].isnull().all()
    assert results_table["querier"][2:].notnull().all()


def test_n_prior_included(tmp_project, demo_data_path, tmpdir):
    argv = f"{demo_data_path} -o {tmp_project} --n-prior-included 2 --prior-seed 535".split()
    _cli_simulate(argv)

    with asr.Project.load(tmp_project, tmpdir).db as db:
        result = db.get_results_table(["label", "querier"])

    prior_included = result["label"] & (result["querier"].isnull())
    assert sum(prior_included) >= 2


def test_n_prior_excluded(tmp_project, demo_data_path, tmpdir):
    argv = f"{demo_data_path} -o {tmp_project} --n-prior-excluded 2 --prior-seed 535".split()
    _cli_simulate(argv)

    with asr.Project.load(tmp_project, tmpdir).db as db:
        result = db.get_results_table(["label", "querier"])

    prior_excluded = ~result["label"] & (result["querier"].isnull())
    assert sum(prior_excluded) >= 2


def test_no_seed(tmpdir, demo_data_path):
    for i in range(10):
        project_fp = Path(tmpdir, f"tmp_state_{i}.asreview")

        argv = (
            f"{demo_data_path} -o {project_fp} -c nb -q max -e tfidf "
            f"--n-prior-excluded 1 --n-prior-included 1 --n-stop 4".split()
        )
        _cli_simulate(argv)
        with asr.Project.load(project_fp, tmpdir).db as db:
            priors = db.get_priors()

    assert len(priors) == 2


@pytest.mark.parametrize("seed", [(535), (165), (42)])
def test_model_and_prior_seed(tmpdir, seed, demo_data_path):
    project1_fp = Path(tmpdir, "tmp_state1.asreview")
    project2_fp = Path(tmpdir, "tmp_state2.asreview")

    # run the simulations
    _cli_simulate(
        f"{demo_data_path} -o {project1_fp} -c rf -e tfidf -q max --prior-seed {seed}"
        f" --seed {seed}"
        f" --n-prior-excluded 1 --n-prior-included 1".split()
    )
    _cli_simulate(
        f"{demo_data_path} -o {project2_fp} -c rf -e tfidf -q max --prior-seed {seed}"
        f" --seed {seed}"
        f" --n-prior-excluded 1 --n-prior-included 1".split()
    )

    # open the results database and extract the priors
    with asr.Project.load(project1_fp, tmpdir).db as db1:
        record_table1 = db1.get_results_table().drop("time", axis=1)

    with asr.Project.load(project2_fp, tmpdir).db as db2:
        record_table2 = db2.get_results_table().drop("time", axis=1)

    assert_frame_equal(record_table1, record_table2)


@pytest.mark.parametrize("model", ["logistic", "nb", "rf", "svm"])
def test_models(model, tmpdir, demo_data_path, tmp_project):
    _cli_simulate(
        f"{demo_data_path} -o {tmp_project} -c {model}".split()
        + "-e tfidf -q max -b balanced".split()
    )

    with asr.Project.load(tmp_project, tmpdir).db as db:
        results = db.get_results_table()

    assert all(results["classifier"][10:] == model)
    assert all(results["balancer"][10:] == "balanced")
    assert all(results["feature_extractor"][10:] == "tfidf")
    assert all(results["querier"][10:] == "max")


def test_no_balancing(tmp_project, demo_data_path, tmpdir):
    argv = f"{demo_data_path} -o {tmp_project} -c nb -q max -e tfidf".split()
    _cli_simulate(argv)

    with asr.Project.load(tmp_project, tmpdir).db as db:
        results_balance_strategies = db.get_results_table()["balancer"]
        last_ranking_balance_strategies = db.get_last_ranking_table()["balancer"]

    assert results_balance_strategies.isnull().all()
    assert last_ranking_balance_strategies.isnull().all()


def test_number_records_found(tmp_project, demo_data_path, tmpdir):
    _cli_simulate(
        f"{demo_data_path} -o {tmp_project} --n-stop 15"
        " --prior-idx 0 9 --seed 535".split()
    )

    with asr.Project.load(tmp_project, tmpdir).db as db:
        assert db.get_results_table("label")["label"].sum() == 9
        assert db.get_results_table("label").shape[0] == 15
        assert db.get_results_table().shape[0] == 15
        assert db.get_results_table()["label"].head(2).sum() == 1


def test_n_stop_min(tmp_project, demo_data_path, tmpdir):
    argv = f"{demo_data_path} -o {tmp_project} --prior-idx 0 9 --seed 535".split()
    _cli_simulate(argv)

    with asr.Project.load(tmp_project, tmpdir).db as db:
        assert db.get_results_table("label")["label"].sum() == 10
        assert len(db.get_results_table("label")) == 51


def test_n_stop_all(tmp_project, tmpdir):
    argv = (
        "synergy:Sep_2021",
        "-o",
        f"{tmp_project}",
        "--n-stop",
        "-1",
        "--prior-idx",
        "116",
        "10",
        "--seed",
        "101",
    )
    _cli_simulate(argv)

    with asr.Project.load(tmp_project, tmpdir).db as db:
        assert db.get_results_table("label")["label"].sum() == 40
        assert len(db.get_results_table("label")) == 271


def test_project_already_exists_error(tmp_project, demo_data_path):
    argv = f"{demo_data_path} -o {tmp_project} --n-stop 100 --seed 535".split()
    _cli_simulate(argv)

    with pytest.raises(ValueError):
        # Simulate 100 queries in two steps of 50.
        argv = f"{demo_data_path} -o {tmp_project} --n-stop 50 --seed 535".split()
        _cli_simulate(argv)


def test_cycle_config(tmpdir, demo_data_path, tmp_project):
    cycle_meta = asr.ActiveLearningCycleData(
        querier="max",
        classifier="nb",
        balancer="balanced",
        feature_extractor="tfidf",
        classifier_param={"alpha": 5},
    )

    fp_cycle = Path(tmpdir, "cycle.json")

    with open(fp_cycle, "w") as f:
        json.dump(asdict(cycle_meta), f)

    _cli_simulate(f"{demo_data_path} -o {tmp_project} --config-file {fp_cycle}".split())

    with asr.Project.load(tmp_project, tmpdir).db as db:
        results = db.get_results_table()

    assert all(results["classifier"][10:] == "nb")
    assert all(results["balancer"][10:] == "balanced")
    assert all(results["feature_extractor"][10:] == "tfidf")
    assert all(results["querier"][10:] == "max")

    # todo save and test for params in simulation


def test_grouped_records(tmp_path, demo_data, tmp_project, tmpdir):
    data_path = tmp_path / "duplicate_data.csv"
    duplicate_demo_data = pd.concat([demo_data, demo_data])
    duplicate_demo_data.to_csv(data_path, index=False)

    argv = f"{data_path} -o {tmp_project} --group-similar-records".split()
    _cli_simulate(argv)

    with asr.Project.load(tmp_project, tmpdir).db as db:
        results = pd.read_sql_query("SELECT * FROM results", db._conn)

    # Every record is duplicated exactly once. So after every labeling action, the data
    # of a single duplicate record is added to the state file as well. The duplicate
    # record will have record_id equal to the normal record_id + 100.
    for i in range(0, len(results), 2):
        pair = results.iloc[i : i + 2]
        ids = sorted(pair["record_id"].to_list())
        assert ids[1] == ids[0] + 100
        assert pair["label"].iloc[0] == pair["label"].iloc[1]



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.pipeline import Pipeline

__all__ = ["Tfidf", "OneHot"]


class TextMerger(TransformerMixin, BaseEstimator):
    """Merge text columns into a single column.

    Merge multiple columns into a single column. This can be useful when
    multiple columns contain text information that should be combined.

    Parameters
    ----------
    columns: list
        List of columns to merge.
    sep: str
        Separator to use when merging the columns.
    """

    def __init__(self, columns, sep=" "):
        self.columns = columns
        self.sep = sep

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X[self.columns].fillna("").apply(lambda x: self.sep.join(x), axis=1)


class Tfidf(Pipeline):
    """TF-IDF feature extraction.

    Based on the sklearn implementation of the TF-IDF feature extraction
    sklearn.feature_extraction.text.TfidfVectorizer.

    Parameters
    ----------
    columns: list, default=["title", "abstract"]
        See TextMerger
    sep: str, default=" "
        See TextMerger
    lowercase: bool, default=True
        See ScikitLearn CountVectorizer
    stop_words: {'english'} or list or None, default=None
        See ScikitLearn CountVectorizer
    token_pattern: str or None, default=r"(?u)\\b\\w\\w+\\b"
        See ScikitLearn CountVectorizer
    ngram_range: tuple (min_n, max_n), default=(1,1)
        See ScikitLearn CountVectorizer
    max_df: float in range [0.0, 1.0] or int, default=1.0
        See ScikitLearn CountVectorizer
    min_df: float in range [0.0, 1.0] or int, default=1
        See ScikitLearn CountVectorizer
    max_features: int, default=None
        See ScikitLearn CountVectorizer
    vocabulary: Mapping or iterable, default=None
        See ScikitLearn CountVectorizer
    binary: bool, default=False
        See ScikitLearn CountVectorizer
    norm: {"l1", "l2"} or None, default="l2"
        See ScikitLearn CountVectorizer
    use_idf: bool, default=True
        See ScikitLearn CountVectorizer
    smooth_idf: bool, default=True
        See ScikitLearn CountVectorizer
    sublinear_tf: bool, default=False
        See ScikitLearn CountVectorizer
    **kwargs: dict
        See ScikitLearn CountVectorizer for additional parameters
    """

    name = "tfidf"
    label = "TF-IDF"

    def __init__(
        self,
        columns=["title", "abstract"],
        sep=" ",
        lowercase=True,
        stop_words=None,
        token_pattern=r"(?u)\b\w\w+\b",
        ngram_range=(1, 1),
        max_df=1.0,
        min_df=1,
        max_features=None,
        vocabulary=None,
        binary=False,
        norm="l2",
        use_idf=True,
        smooth_idf=True,
        sublinear_tf=False,
        **kwargs,
    ):
        self.columns = columns
        self.sep = sep
        self.lowercase = lowercase
        self.stop_words = stop_words
        self.token_pattern = token_pattern
        self.ngram_range = tuple(ngram_range)
        self.max_df = max_df
        self.min_df = min_df
        self.max_features = max_features
        self.vocabulary = vocabulary
        self.binary = binary
        self.norm = norm
        self.use_idf = use_idf
        self.smooth_idf = smooth_idf
        self.sublinear_tf = sublinear_tf
        super().__init__(
            [
                ("text_merger", TextMerger(columns=self.columns, sep=self.sep)),
                (
                    "tfidf",
                    TfidfVectorizer(
                        lowercase=self.lowercase,
                        stop_words=self.stop_words,
                        token_pattern=self.token_pattern,
                        ngram_range=self.ngram_range,
                        max_df=self.max_df,
                        min_df=self.min_df,
                        max_features=self.max_features,
                        vocabulary=self.vocabulary,
                        binary=self.binary,
                        norm=self.norm,
                        use_idf=self.use_idf,
                        smooth_idf=self.smooth_idf,
                        sublinear_tf=self.sublinear_tf,
                        **kwargs,
                    ),
                ),
            ]
        )


class OneHot(Pipeline):
    """One-hot feature extraction.

    Based on the sklearn implementation of the one-hot feature extraction
    sklearn.feature_extraction.text.CountVectorizer with binary=True.

    Parameters
    ----------
    columns: list, default=["title", "abstract"]
        See TextMerger
    sep: str, default=" "
        See TextMerger
    lowercase: bool, default=True
        See ScikitLearn CountVectorizer
    stop_words: {'english'} or list or None, default=None
        See ScikitLearn CountVectorizer
    token_pattern: str or None, default=r"(?u)\\b\\w\\w+\\b"
        See ScikitLearn CountVectorizer
    ngram_range: tuple (min_n, max_n), default=(1,1)
        See ScikitLearn CountVectorizer
    max_df: float in range [0.0, 1.0] or int, default=1.0
        See ScikitLearn CountVectorizer
    min_df: float in range [0.0, 1.0] or int, default=1
        See ScikitLearn CountVectorizer
    max_features: int, default=None
        See ScikitLearn CountVectorizer
    vocabulary: Mapping or iterable, default=None
        See ScikitLearn CountVectorizer
    **kwargs: dict
        See ScikitLearn CountVectorizer for additional parameters
    """

    name = "onehot"
    label = "OneHot"

    def __init__(
        self,
        columns=["title", "abstract"],
        sep=" ",
        lowercase=True,
        stop_words=None,
        token_pattern=r"(?u)\b\w\w+\b",
        ngram_range=(1, 1),
        max_df=1.0,
        min_df=1,
        max_features=None,
        vocabulary=None,
        **kwargs,
    ):
        self.columns = columns
        self.sep = sep
        self.lowercase = lowercase
        self.token_pattern = token_pattern
        self.stop_words = stop_words
        self.max_df = max_df
        self.min_df = min_df
        self.max_features = max_features
        self.ngram_range = tuple(ngram_range)
        self.vocabulary = vocabulary
        super().__init__(
            [
                ("text_merger", TextMerger(columns=self.columns, sep=self.sep)),
                (
                    "onehot",
                    CountVectorizer(
                        lowercase=self.lowercase,
                        token_pattern=self.token_pattern,
                        stop_words=self.stop_words,
                        max_df=self.max_df,
                        min_df=self.min_df,
                        max_features=self.max_features,
                        ngram_range=self.ngram_range,
                        vocabulary=self.vocabulary,
                        binary=True,
                        **kwargs,
                    ),
                ),
            ]
        )



import sqlite3
from pathlib import Path

import pandas as pd
import pytest
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Mapped

from asreview.data.loader import load_dataset
from asreview.data.loader import load_records
from asreview.data.record import Base
from asreview.data.record import Record
from asreview.data.utils import _clean_text
from asreview.data.utils import identify_groups
from asreview.data.utils import identify_record_groups
from asreview.database.store import DataStore
from asreview.project.api import Project


@pytest.fixture
def records():
    data_fp = Path("tests", "demo_data", "generic.csv")
    return load_records(data_fp, dataset_id="foo")


@pytest.fixture
def store(tmpdir):
    fp = tmpdir / Project.PATH_DB
    store = DataStore(fp)
    store.create_tables()
    return store


@pytest.fixture
def store_with_data(store, records):
    store.add_records(records)
    return store


def test_create_tables(tmpdir):
    fp = tmpdir / Project.PATH_DB
    store = DataStore(fp)
    store.create_tables()
    con = sqlite3.connect(fp)
    tables = (
        con.cursor()
        .execute("SELECT name FROM sqlite_master WHERE type='table'")
        .fetchone()
    )
    assert tables == ("record",)


def test_add_dataset(store, records):
    store.add_records(records)
    con = sqlite3.connect(store.fp)
    df = pd.read_sql("SELECT * FROM record", con)
    assert df["dataset_id"].eq("foo").all()


def test_record_id_start_at_zero(store):
    store.add_records(
        [
            Record(dataset_id="foo", dataset_row=5),
            Record(dataset_id="foo", dataset_row=6),
        ]
    )
    assert store["record_id"].to_list() == [0, 1]


def test_is_empty(store, records):
    assert store.is_empty()
    store.add_records(records)
    assert not store.is_empty()


def test_get_records(store, records):
    store.add_records(records)
    row_number = 1
    record = store.get_records(row_number)
    assert isinstance(record, Record)
    assert record.record_id == row_number
    all_records = store.get_records()
    assert len(all_records) == len(records)


def test_get_df(store_with_data):
    output = store_with_data.get_df()
    assert isinstance(output, pd.DataFrame)
    assert set(output.columns) == set(Record.get_columns())


def test_len(store, records):
    assert len(store) == 0
    store.add_records(records)
    assert len(store) == len(records)


def test_get_column(store_with_data, records):
    abstracts = store_with_data["abstract"]
    assert isinstance(abstracts, pd.Series)
    abstracts = store_with_data[["abstract"]]
    assert isinstance(abstracts, pd.DataFrame)
    assert abstracts.columns == ["abstract"]
    for i in range(len(records)):
        assert abstracts.loc[i, "abstract"] == records[i].abstract

    data = store_with_data[["title", "record_id"]]
    assert isinstance(data, pd.DataFrame)
    assert list(data.columns) == ["title", "record_id"]


def test_custom_record(tmpdir):
    class CustomRecord(Base):
        __tablename__ = "custom"
        foo: Mapped[str]

    fp = Path(tmpdir, Project.PATH_DB)
    data_store = DataStore(fp, CustomRecord)
    data_store.create_tables()


def test_close_store(tmpdir):
    fp = Path(tmpdir, Project.PATH_DB)
    data_store = DataStore(fp)
    data_store.create_tables()
    fp.unlink()


def test_dataset_row_id_unique(store):
    records = [
        Record(dataset_row=1, dataset_id="foo"),
        Record(dataset_row=1, dataset_id="foo"),
    ]
    with pytest.raises(IntegrityError):
        store.add_records(records)


def test_author_validation(store):
    records = [
        Record(dataset_id="foo", dataset_row=1, authors=None),
        Record(dataset_id="foo", dataset_row=2, authors=["Foo", "Bar"]),
    ]
    store.add_records(records)
    with pytest.raises(ValueError):
        store.add_records([Record(dataset_id="foo", dataset_row=3, authors="Foo;Bar")])


def test_keyword_validation(store):
    records = [
        Record(dataset_id="foo", dataset_row=1, keywords=None),
        Record(dataset_id="foo", dataset_row=2, keywords=["Foo", "Bar"]),
    ]
    store.add_records(records)
    with pytest.raises(ValueError):
        store.add_records([Record(dataset_id="foo", dataset_row=3, keywords="Foo;Bar")])


def test_label_validation(store):
    records = [
        Record(dataset_id="foo", dataset_row=1, included=None),
        Record(dataset_id="foo", dataset_row=2, included=1),
        Record(dataset_id="foo", dataset_row=3, included=0),
    ]
    store.add_records(records)
    with pytest.raises(ValueError):
        store.add_records([Record(dataset_id="foo", dataset_row=4, included="1")])


def test_delete_record(store):
    record = Record(dataset_id="foo", dataset_row=1)
    store.add_records([record])
    assert len(store) == 1
    store.delete_record(record.record_id)
    assert len(store) == 0


def test_load_dataset_no_abstracts(tmpdir):
    test_fp = tmpdir / "no_abstracts.csv"
    df = pd.DataFrame(
        {
            "title": ["Title 1", "Title 2"],
            "authors": ["Author A", "Author B"],
            "abstract": [None, None],
        }
    )
    df.to_csv(test_fp, index=False)
    db = load_dataset(test_fp, dataset_id="foo")

    assert db.input.get_df()["abstract"].replace("", None).isnull().all()


@pytest.mark.parametrize(
    "file_name,n_lines",
    [
        ("_baseline.ris", 100),
        ("embase.csv", 6),
        ("embase_newpage.csv", 6),
        ("embase.ris", 6),
        ("generic.csv", 2),
        ("generic_semicolon.csv", 2),
        ("generic_tab.csv", 2),
        ("generic_tab.tab", 2),
        ("generic_tab.tsv", 2),
        ("generic_labels.csv", 6),
        ("pubmed_zotero.ris", 6),
        ("pubmed_endnote.txt", 6),
        ("scopus.ris", 6),
        ("ovid_zotero.ris", 6),
        ("proquest.ris", 6),
        ("web_of_science.txt", 10),
    ],
)
def test_load_dataset(file_name, n_lines):
    fp = Path("tests", "demo_data", file_name)
    with load_dataset(fp, dataset_id=file_name) as db:
        assert len(db.input) == n_lines


@pytest.mark.internet_required
def test_load_dataset_from_url(osf_fg93a_path):
    store = load_dataset(osf_fg93a_path, dataset_id="osf_fg93a")
    assert len(store.input) == 38


def test_dataset_with_record_ids():
    fp = Path("tests", "demo_data", "record_id.csv")
    db = load_dataset(fp)
    record_ids = db.input["record_id"]
    assert record_ids.to_list() == list(range(len(db.input)))


def test_load_faulty_year_dataset():
    # The dataset contains one record with the value '10 15' as publication year.
    with pytest.raises(ValueError):
        load_dataset(Path("tests", "demo_data", "faulty_year.ris"))


@pytest.mark.parametrize(
    "groups, normalized_chain",
    [
        ([(0, 0)], {0: None, 1: None, 2: None, 3: None}),
        ([(0, 0), (0, 1)], {0: None, 1: 0, 2: None, 3: None}),
        ([(0, 0), (0, 2), (1, 1)], {0: None, 1: None, 2: 0, 3: None}),
        ([(3, 0), (1, 1), (1, 2), (3, 3)], {0: None, 1: None, 2: 1, 3: 0}),
    ],
)
def test_set_groups(store, groups, normalized_chain):
    records = [Record(dataset_row=idx, dataset_id="foo") for idx in range(4)]
    store.add_records(records)
    store.set_groups(groups)
    stored_duplicate_chain = {
        record.record_id: record.duplicate_of for record in store.get_records()
    }
    assert stored_duplicate_chain == normalized_chain


@pytest.mark.parametrize(
    "groups",
    [
        [(0, 0), (1, 1), (2, 2), (3, 3)],
        [(0, 0), (0, 1), (2, 2), (3, 3)],
        [(0, 0), (0, 2), (0, 3), (1, 1)],
        [(0, 0), (0, 2), (1, 1), (1, 3)],
        [(0, 0), (0, 1), (0, 2), (0, 3)],
    ],
)
def test_get_groups(store, groups):
    records = [Record(dataset_row=idx, dataset_id="foo") for idx in range(len(groups))]
    store.add_records(records)
    store.set_groups(groups)
    stored_groups = store.get_groups()
    assert stored_groups == groups

    for group_id, record_id in groups:
        group_members = set((g_id, r_id) for (g_id, r_id) in groups if g_id == group_id)
        assert group_members == set(store.get_groups(record_id=record_id))


@pytest.mark.parametrize(
    "groups, result",
    [
        ([], [(0, 0), (1, 1), (2, 2), (3, 3)]),
        ([(0, 0), (0, 1)], [(0, 0), (0, 1), (2, 2), (3, 3)]),
        ([(0, 0), (0, 2), (0, 3)], [(0, 0), (0, 2), (0, 3), (1, 1)]),
    ],
)
def test_get_groups_partial_data(store, groups, result):
    records = [Record(dataset_row=idx, dataset_id="foo") for idx in range(4)]
    store.add_records(records)
    store.set_groups(groups)
    stored_groups = store.get_groups()
    assert stored_groups == result

    for group_id, record_id in result:
        group_members = set((g_id, r_id) for (g_id, r_id) in result if g_id == group_id)
        assert group_members == set(store.get_groups(record_id=record_id))


@pytest.mark.parametrize(
    "input_data,expected",
    [
        # No duplicates
        (["a", "b", "c"], [(0, 0), (1, 1), (2, 2)]),
        # Simple duplicate
        (["a", "b", "a"], [(0, 0), (1, 1), (0, 2)]),
        # Multiple duplicates of same element
        (["x", "x", "x"], [(0, 0), (0, 1), (0, 2)]),
        # Mix of unique and duplicate
        (["a", "b", "a", "c", "b"], [(0, 0), (1, 1), (0, 2), (3, 3), (1, 4)]),
        # Integers instead of strings
        ([1, 2, 3, 1, 2], [(0, 0), (1, 1), (2, 2), (0, 3), (1, 4)]),
        # Empty iterable
        ([], []),
    ],
)
def test_identify_groups(input_data, expected):
    assert identify_groups(input_data) == expected


@pytest.mark.parametrize(
    "records, record_ids, feature_extractors, expected",
    [
        # Case 1: identical titles <E2><86><92> grouped
        (
            [
                Record("ds1", 1, title="A"),
                Record("ds1", 2, title="A"),
                Record("ds1", 3, title="B"),
            ],
            [0, 1, 2],
            [lambda r: r.title],
            [(0, 0), (0, 1), (2, 2)],
        ),
        # Case 2: use dataset_id and dataset_row <E2><86><92> all unique
        (
            [
                Record("ds1", 1, title="A"),
                Record("ds1", 2, title="A"),
                Record("ds1", 3, title="A"),
            ],
            [4, 2, 7],
            [lambda r: r.dataset_id, lambda r: r.dataset_row],
            # All unique <E2><86><92> group id = record id
            [(4, 4), (2, 2), (7, 7)],
        ),
        # Case 3: two different groups by DOI
        (
            [
                Record("ds1", 1, doi="x"),
                Record("ds1", 2, doi="y"),
                Record("ds1", 3, doi="x"),
            ],
            [3, 4, 5],
            [lambda r: r.doi],
            # First and third same group
            [(3, 3), (4, 4), (3, 5)],
        ),
    ],
)
def test_identify_record_groups(records, record_ids, feature_extractors, expected):
    for record, record_id in zip(records, record_ids):
        record.record_id = record_id
    result = identify_record_groups(records, feature_extractors)
    assert set(result) == set(expected)


def test_load_dataset_grouped(tmpdir):
    # Create a ris file that contains another dataset twice.
    with open(Path("tests", "demo_data", "pubmed_zotero.ris")) as f:
        file_text = f.read()
    duplicate_fp = Path(tmpdir, "duplicate_test.ris")
    # duplicate_fp = Path("tests", "demo_data", "duplicated_dataset.ris")
    with open(duplicate_fp, "w") as f:
        f.write(file_text + "\n\n" + file_text)

    db = load_dataset(duplicate_fp)
    # The original dataset has 6 records, so this one has 12.
    assert len(db.input) == 12
    # There should be six groups, each containing the original record and the duplicated
    # record six row lower.
    groups = db.input.get_groups()
    assert set(groups) == set((i, i) for i in range(6)).union(
        set((i, i + 6) for i in range(6))
    )


@pytest.mark.parametrize(
    "text,expected",
    [
        # Empty / falsy inputs
        ("", ""),
        (None, ""),
        # Lowercasing
        ("Machine Learning", "machinelearning"),
        # Whitespace removal
        ("  hello   world  ", "helloworld"),
        ("hello\tworld\n", "helloworld"),
        # Punctuation removal
        ("self-report", "selfreport"),
        ("title: a study.", "titleastudy"),
        # Accented characters normalized via NFKD
        ("café", "cafe"),
        ("naïve résumé", "naiveresume"),
        # Unicode dashes normalized (em-dash, en-dash)
        ("self\u2014report", "selfreport"),
        ("self\u2013report", "selfreport"),
        # Ligatures expanded via NFKD
        ("\ufb01nding", "finding"),
        # Mixed
        ("  Héllo,  Wörld!  ", "helloworld"),
    ],
)
def test_clean_text(text, expected):
    assert _clean_text(text) == expected


def test_identify_record_groups_default_extractors():
    """Test that the default extractors detect duplicates despite whitespace,
    case, punctuation, and accent differences."""
    records = [
        Record("ds1", 0, title="Machine Learning", abstract="An abstract."),
        Record("ds1", 1, title="machine  learning", abstract="an abstract"),
        Record("ds1", 2, title="MACHINE-LEARNING", abstract="An Abstract!"),
        Record("ds1", 3, title="café study", abstract="some abstract"),
        Record("ds1", 4, title="cafe study", abstract="some abstract"),
        Record("ds1", 5, title="unique title", abstract="unique abstract"),
    ]
    for i, record in enumerate(records):
        record.record_id = i

    result = identify_record_groups(records)

    # First three records should be in one group (group_id=0).
    # "café study" and "cafe study" should be in another group (group_id=3).
    # "unique title" should be its own group.
    assert set(result) == {(0, 0), (0, 1), (0, 2), (3, 3), (3, 4), (5, 5)}



import os
import sqlite3

import pytest
from pathlib import Path

import asreview as asr
from asreview.models.balancers import Balanced
from asreview.models.classifiers import SVM


def test_init_project_folder(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    project = asr.Project.create(project_path)

    assert Path(project_path, project.PATH_CONFIG).is_file()
    assert project.data_dir.is_dir()
    assert Path(project_path, project.PATH_FEATURE_MATRICES).is_dir()
    assert project.config["id"] == "test"


def test_init_project_already_exists(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    asr.Project.create(project_path)
    with pytest.raises(ValueError):
        asr.Project.create(project_path)


def test_project_load(asreview_test_project_path, tmpdir):
    project = asr.Project.load(asreview_test_project_path, tmpdir)

    assert project.db_path.exists()
    assert Path(project.project_path, project.PATH_CONFIG).exists()
    assert isinstance(project.db, asr.Database)
    project.db._is_valid()


def test_project_load_unknown_classifier(tmpdir):
    project_fp = Path(
        "tests", "asreview_files", "asreview-demo-project-invalid-classifier.asreview"
    )

    with pytest.warns(
        UserWarning, match="'idontknow' not found in group models.classifiers"
    ):
        project = asr.Project.load(project_fp, tmpdir, reset_model_if_not_found=True)
    cycle = asr.ActiveLearningCycle.from_meta(
        asr.ActiveLearningCycleData(**project.get_model_config())
    )

    assert project.config["review"]["model"]["name"].startswith("elas_u")
    assert isinstance(cycle.classifier, SVM)


def test_al_cycle_state(asreview_test_project, tmpdir):
    cycle = asr.ActiveLearningCycle.from_meta(
        asr.ActiveLearningCycleData(**asreview_test_project.get_model_config())
    )

    assert asreview_test_project.config["review"]["model"]["name"].startswith("elas_u")
    assert isinstance(cycle.balancer, Balanced)


def test_project_closes_db_on_exception(tmpdir):
    """Test that Project closes database connection even when exception occurs."""
    project_path = Path(tmpdir, "test.asreview")
    asr.Project.create(project_path)

    with pytest.raises(RuntimeError):
        with asr.Project(project_path) as project:
            # access db to trigger the cached_property
            conn = project.db._conn
            raise RuntimeError("simulated failure")

    # connection should be closed
    with pytest.raises(
        sqlite3.ProgrammingError, match="Cannot operate on a closed database"
    ):
        conn.execute("SELECT 1")

    # db file should be unlocked (this is what fails on Windows with leaked connections)
    os.remove(project.db_path)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from asreview.metrics import loss
from asreview.metrics import ndcg


@pytest.mark.parametrize(
    "labels, expected_value",
    [
        ([1, 0], 0),
        ([0, 1], 1),
        ([1, 1, 0, 0, 0], 0),
        ([0, 0, 0, 1, 1], 1),
        ([1, 0, 1], 0.5),
        (
            [
                0,
                1,
                0,
                0,
                0,
                1,
                0,
                1,
                0,
                1,
                0,
                0,
                1,
                1,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                1,
                0,
                1,
                0,
                0,
                1,
                1,
            ],
            0.5583333333333333,
        ),
    ],
)
def test_loss_value_function(labels, expected_value):
    loss_value = loss(labels)
    assert_almost_equal(loss_value, expected_value)


@pytest.mark.parametrize(
    "labels",
    [[0, 0, 0], [0], [1]],
)
def test_loss_value_error_cases(labels):
    with pytest.raises(ValueError):
        loss(labels)


def test_random_loss_values():
    np.random.seed(42)
    for _ in range(100):
        length = np.random.randint(2, 100)
        labels = np.random.randint(0, 2, length)

        # Ensure labels are not all 0 or all 1
        if np.all(labels == 0) or np.all(labels == 1):
            labels[np.random.randint(0, length)] = 1 - labels[0]

        loss_value = loss(labels)
        assert 0 <= loss_value <= 1


@pytest.mark.parametrize(
    "labels, expected_value",
    [
        ([1, 0], 1),
        ([0, 1], 0.6309297535714575),
        ([1, 1, 0, 0, 0], 1),
        ([0, 0, 0, 1, 1], 0.5012658353418871),
        ([1, 0, 1], 0.9197207891481876),
        (
            [
                0,
                1,
                0,
                0,
                0,
                1,
                0,
                1,
                0,
                1,
                0,
                0,
                1,
                1,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                1,
                0,
                1,
                0,
                0,
                1,
                1,
            ],
            0.6627901128094049,
        ),
    ],
)
def test_ndcg_value_function(labels, expected_value):
    loss_value = ndcg(labels)
    assert_almost_equal(loss_value, expected_value)


@pytest.mark.parametrize(
    "labels",
    [[0, 0, 0], [0], [1]],
)
def test_lndcg_value_error_cases(labels):
    with pytest.raises(ValueError):
        ndcg(labels)


def test_random_ndcg_values():
    np.random.seed(42)
    for _ in range(100):
        length = np.random.randint(2, 100)
        labels = np.random.randint(0, 2, length)

        # Ensure labels are not all 0 or all 1
        if np.all(labels == 0) or np.all(labels == 1):
            labels[np.random.randint(0, length)] = 1 - labels[0]

        loss_value = ndcg(labels)
        assert 0 <= loss_value <= 1



import json
import os
import sqlite3
from dataclasses import asdict
from pathlib import Path
from uuid import uuid4

import pandas

from asreview._version import __version__
from asreview.data.loader import _get_reader
from asreview.models.models import get_ai_config
from asreview.project.migration.legacy_v2.record import Record as RecordV2
from asreview.project.migration.legacy_v2.sqlstate import SQLiteState
from asreview.project.migration.legacy_v2.store import DataStore


def _project_config_converter_v1_v2(project_json):
    # Convert the project config from version 1 to version 2.
    # This is a helper function for the main migration function.

    project_json["version"] = __version__
    project_json["project_file_version"] = 2
    project_json.pop("datetimeCreated", None)
    project_json.pop("description", None)
    project_json.pop("authors", None)

    if project_json["mode"] == "explore":
        project_json["mode"] = "oracle"

    for i, _ in enumerate(project_json["reviews"]):
        del project_json["reviews"][i]["start_time"]
        try:
            del project_json["reviews"][i]["end_time"]
        except KeyError:
            pass

        # fix projects with training errors
        if project_json["reviews"][i]["status"] == "error":
            project_json["reviews"][i]["status"] = "review"

    return project_json


def _project_data_converter_v1_v2(project_config, project_folder):
    fp_data = Path(project_folder, "data", project_config["dataset_path"])

    dataset_id = uuid4().hex
    reader = _get_reader(fp_data)
    records = reader.read_records(fp_data, dataset_id=dataset_id, record_cls=RecordV2)

    data_store = DataStore(Path(project_folder, "data_store.db"))
    data_store.create_tables()
    data_store.add_records(records=records)

    project_config["datasets"] = [{"id": dataset_id, "name": fp_data.name}]
    del project_config["dataset_path"]


def _project_state_converter_v1_v2(review_path):
    sqlstate = SQLiteState(Path(review_path, "results.db"))
    sqlstate.create_tables()

    conn = sqlite3.connect(Path(review_path, "results.sql"))

    df_results = pandas.read_sql_query(
        "SELECT * FROM results WHERE label is not NULL", conn
    ).rename(
        columns={
            "notes": "note",
            "query_strategy": "querier",
            "balance_strategy": "balancer",
            "feature_extraction": "feature_extractor",
        }
    )
    df_results["tags"] = "[]"
    df_results["user_id"] = None
    df_results["time"] = (
        pandas.to_datetime(df_results["labeling_time"]).astype("int64")
        // int(1e3)
        / int(1e6)
    )
    df_results["querier"] = df_results["querier"].replace("prior", None)
    del df_results["labeling_time"]
    sqlstate._replace_results_from_df(df_results)

    df_last_ranking = pandas.read_sql_query("SELECT * FROM last_ranking", conn)

    if not df_last_ranking.empty:
        df_last_ranking = df_last_ranking.rename(
            columns={
                "query_strategy": "querier",
                "balance_strategy": "balancer",
                "feature_extraction": "feature_extractor",
            }
        )
        df_last_ranking["time"] = (
            pandas.to_datetime(df_last_ranking["time"]).astype("int64")
            // int(1e3)
            / int(1e6)
        )

        sqlstate._replace_last_ranking_from_df(df_last_ranking)

    df_decision_changes = pandas.read_sql_query("SELECT * FROM decision_changes", conn)
    df_decision_changes["time"] = (
        pandas.to_datetime(df_decision_changes["time"]).astype("int64")
        // int(1e3)
        / int(1e6)
    )
    df_decision_changes.rename(
        columns={
            "new_label": "label",
        },
        inplace=True,
    )
    # add column user_id
    df_decision_changes["user_id"] = None

    df_decision_changes.to_sql(
        "decision_changes", sqlstate._conn, if_exists="replace", index=False
    )

    conn.close()
    sqlstate.close()

    os.unlink(Path(review_path, "results.sql"))


def _project_model_settings_converter_v1_v2(fp_cycle_metadata):
    with open(fp_cycle_metadata, "w") as f:
        default_model = get_ai_config()
        json.dump(
            {
                "name": default_model["name"],
                "current_value": asdict(default_model["value"]),
            },
            f,
        )


def _migrate(folder):
    """Migrate a project from version 1 to version 2.

    Parameters
    ----------
    folder: str
        The folder of the project to migrate

    Returns
    -------
    None
    """

    fp_project_config = Path(folder, "project.json")

    # update the project config file
    with open(fp_project_config) as f:
        project_config = json.load(f)

    project_config = _project_config_converter_v1_v2(project_config)

    # remove the project.json.lock file
    try:
        os.unlink(Path(folder, "project.json.lock"))
    except FileNotFoundError:
        pass

    # remove all pickle files
    for file in Path(folder).glob("*.pickle"):
        file.unlink()

    # update the data file
    _project_data_converter_v1_v2(project_config, folder)

    with open(fp_project_config, "w") as f:
        json.dump(project_config, f)

    # update the state file
    for review in project_config["reviews"]:
        _project_state_converter_v1_v2(Path(folder, "reviews", review["id"]))

        # Update the model settings file
        _project_model_settings_converter_v1_v2(
            Path(folder, "reviews", review["id"], "settings_metadata.json"),
        )



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# based on https://github.com/pypa/sampleproject - MIT License

# Always prefer setuptools over distutils
import platform
import subprocess
from pathlib import Path

from setuptools import Command
from setuptools import setup


class CompileAssets(Command):
    """
    Compile and build the frontend assets using npm and webpack.

    Registered as cmdclass in setup() so it can be called with
    ``python setup.py compile_assets``.
    """

    description = "Compile and build the frontend assets"
    user_options = []

    def initialize_options(self):
        """Set default values for options."""

    def finalize_options(self):
        """Set final values for options."""

    def run(self):
        """Run a command to compile and build assets."""

        path_webapp = Path(__file__).parent / "asreview" / "webapp"

        subprocess.check_call(
            ["npm", "install"],
            cwd=str(path_webapp),
            shell=(platform.system() == "Windows"),
        )
        subprocess.check_call(
            ["npm", "run-script", "build"],
            cwd=str(path_webapp),
            shell=(platform.system() == "Windows"),
        )


def get_cmdclass():
    cmdclass = {}
    cmdclass["compile_assets"] = CompileAssets
    return cmdclass


setup(
    cmdclass=get_cmdclass(),
)



import io
import json
import random
import re
import zipfile
from pathlib import Path
from typing import Union

from flask import current_app

from asreview.webapp.utils import asreview_path


def custom_remote_auth_headers(
    remote_user="foo",
    affiliation="UU",
    email="foo@dev.bar",
    name="Foo Bar",
    secret="secret",
):
    return {
        "environ_base": {
            "REMOTE_USER": remote_user,
            "REMOTE_USER_EMAIL": email,
            "REMOTE_USER_AFFILIATION": affiliation,
            "REMOTE_AUTH_SECRET": secret,
        },
        "headers": {},
    }


def get_project_id(project):
    """Get a project id from either a Project model
       (authenticated app) or an asr.Project
    object
       (unauthenticated app)."""
    if not current_app.config.get("AUTHENTICATION"):
        return project.config["id"]

    return project.project_id


def read_project_file(project):
    """Loads the data from the project.json file."""
    id = get_project_id(project)
    with open(asreview_path() / id / "project.json") as f:
        data = json.load(f)
        return data


def manipulate_project_file(project, key, value):
    """Updates key value pairs in the project.json file."""
    id = get_project_id(project)
    data = read_project_file(project)
    data[key] = value
    with open(asreview_path() / id / "project.json", "w+") as f:
        json.dump(data, f)
        return True
    return False


def _extract_stem(path: Union[str, Path]):
    """Extracts a stem from a path or URL containing a filename."""
    return Path(re.split(":|/", str(path))[-1]).stem


def extract_filename_stem(upload_data):
    """Helper function to get the stem part of a filename from a
    Path or URL contaning a filename."""
    # upload data is a dict with a single key value pair
    value = list(upload_data.values())[0]
    # split this value on either / or :
    return _extract_stem(value)


def choose_project_algorithms():
    """Randomly chooses a model plus the appropriate feature
    extraction, query strategy and balance strategy."""
    classifier = random.choice(["svm", "nb", "logistic"])
    feature_extractor = random.choice(["tfidf"])
    data = {
        "name": None,
        "current_value": {
            "classifier": classifier,
            "feature_extractor": feature_extractor,
            "querier": random.choice(
                ["max", "max_random", "max_uncertainty", "random", "uncertainty"]
            ),
            "balancer": random.choice(["balanced", None]),
        },
    }
    return data


def get_folders_in_asreview_path():
    """This function returns the amount of folders located
    in the asreview folder."""
    return [f for f in asreview_path().glob("*") if f.is_dir()]


def get_zip_file_names(flask_response_data):
    asreview_project_zip = io.BytesIO(flask_response_data)

    zip_file = zipfile.ZipFile(asreview_project_zip)
    return zip_file.namelist()



import numpy as np
import pytest
from sklearn.utils import compute_sample_weight
from sklearn.utils.estimator_checks import check_estimator

from asreview.extensions import extensions
from asreview.extensions import load_extension
from asreview.models.balancers import Balanced


def test_balancers():
    assert len(extensions("models.balancers")) >= 1


@pytest.mark.parametrize("balancer", extensions("models.balancers"))
def test_balance_model_name(balancer):
    model = load_extension("models.balancers", balancer.name)()
    assert model.name == balancer.name


@pytest.mark.parametrize("balancer", extensions("models.balancers"))
def test_balance_model_param(balancer):
    model = load_extension("models.balancers", balancer.name)()
    assert isinstance(model.get_params(), dict)


@pytest.mark.parametrize("balancer", extensions("models.balancers"))
@pytest.mark.skip(reason="Check estimator is not working for balance models.")
def test_balance_check_estimator(balancer):
    model = load_extension("models.balancers", balancer.name)()
    assert check_estimator(model)


def test_balanced():
    y = np.random.randint(0, 2, 100)

    balancer = Balanced()
    # Check that ratio=1.0 gives the same as the 'balanced' setting of
    # `compute_sample_weight`.
    assert np.allclose(
        compute_sample_weight("balanced", y), balancer.compute_sample_weight(y)
    )

    for ratio in [0.25, 1, 5, 10]:
        balancer = Balanced(ratio=ratio)
        sample_weight = balancer.compute_sample_weight(y)
        # Check that the sample weight is normalized to the length of y.
        assert round(sum(sample_weight)) == len(y)
        # Check that (total weight 1's) / (total weight 0's) is equal to the ratio.
        assert sum(sample_weight[y == 1]) / sum(sample_weight[y == 0]) - ratio < 10 ** (
            -5
        )



from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import DateTime
from sqlalchemy.orm import declarative_base
from datetime import datetime
from datetime import timezone

Base = declarative_base()


class ProjectQueueModel(Base):
    """Queue model"""

    __tablename__ = "queue"
    id = Column(Integer, primary_key=True)
    project_id = Column(String(250), nullable=False, unique=True)
    simulation = Column(Boolean, nullable=False, default=False)
    created_at = Column(
        DateTime, nullable=False, default=lambda: datetime.now(timezone.utc)
    )



MODEL_METADATA = {
    "balancers": {
        "balanced": {
            "label": "Balanced Sample Weight",
        },
    },
    "classifiers": {
        "rf": {
            "label": "Random forest",
        },
        "logistic": {
            "label": "Logistic regression",
        },
        "svm": {
            "label": "Support vector machine",
        },
        "nb": {
            "label": "Naive Bayes",
        },
    },
    "feature_extractors": {
        "tfidf": {
            "label": "TF-IDF",
        },
        "onehot": {
            "label": "OneHot",
        },
    },
    "queriers": {
        "random": {
            "label": "Random",
        },
        "uncertainty": {
            "label": "Uncertainty",
        },
        "max": {
            "label": "Maximum",
        },
        "max_uncertainty": {
            "label": "Mixed (95% Maximum and 5% Uncertainty)",
        },
        "max_random": {
            "label": "Mixed (95% Maximum and 5% Random)",
        },
        "top_down": {
            "label": "Top-down",
        },
    },
}



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
from pathlib import Path

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from flask import Flask
from flask import redirect
from flask import request
from flask import send_from_directory
from flask.templating import render_template
from flask_cors import CORS
from flask_login import LoginManager
from flask_login import current_user
from werkzeug.exceptions import InternalServerError

from asreview import __version__ as asreview_version
from asreview.webapp import DB
from asreview.webapp._api import admin
from asreview.webapp._api import auth
from asreview.webapp._api import projects
from asreview.webapp._api import team
from asreview.webapp._authentication.decorators import login_remote_user
from asreview.webapp._authentication.models import User
from asreview.webapp._authentication.oauth_handler import OAuthHandler
from asreview.webapp._authentication.remote_user_handler import RemoteUserHandler
from asreview.webapp._entry_points.migrate import MigrationTool
from asreview.webapp.utils import asreview_path


def create_app(**config_vars):
    """Create a new ASReview webapp.

    For use with WSGI servers, such as gunicorn:

        > gunicorn -w 4 -b "127.0.0.1:5006" "asreview.webapp.app:create_app()"

    Returns
    -------
    Flask:
        The ASReview webapp.
    """

    app = Flask(
        __name__,
        instance_relative_config=True,
        static_folder="build/static",
        template_folder="build",
    )

    app.config.from_prefixed_env("ASREVIEW_LAB")
    app.config.from_mapping(**{k.upper(): v for k, v in config_vars.items()})
    if config_fp := app.config.get("CONFIG_PATH", None):
        app.config.from_file(Path(config_fp).absolute(), load=tomllib.load, text=False)

    # if there are no cors and config is in debug mode, add default cors
    if app.debug and not app.config.get("CORS_ORIGINS", None):
        app.config["CORS_ORIGINS"] = ["http://localhost:3000", "http://127.0.0.1:3000"]

    CORS(app, supports_credentials=True, expose_headers=["Content-Disposition"])

    with app.app_context():
        app.register_blueprint(projects.bp)

    if app.config.get("AUTHENTICATION", True):
        # Login Manager
        login_manager = LoginManager(app)
        login_manager.init_app(app)
        login_manager.session_protection = "strong"

        # set authentication to True when no auth. arguments are provided.
        app.config["AUTHENTICATION"] = True
        app.config["LOGIN_DURATION"] = int(app.config.get("LOGIN_DURATION", 31))

        # Register a callback function for current_user.
        @login_manager.user_loader
        def load_user(user_id):
            return User.query.get(int(user_id))

        if not app.config.get("SQLALCHEMY_DATABASE_URI", None):
            env = "development" if app.debug else "production"
            env = "test" if app.testing else env
            uri = os.path.join(asreview_path(), f"asreview.{env}.sqlite")
            app.config["SQLALCHEMY_DATABASE_URI"] = f"sqlite:///{uri}"

        # initialize app for SQLAlchemy
        DB.init_app(app)

        with app.app_context():
            # create tables in case they don't exist
            DB.create_all()

            # Run database migrations to ensure schema is up to date
            try:
                migration_tool = MigrationTool()
                migration_tool.migrate_database(
                    db_uri=app.config.get("SQLALCHEMY_DATABASE_URI")
                )
            except Exception as e:
                logging.warning(f"Database migration encountered an issue: {e}")
                logging.warning("Continuing with server startup...")

        # authentication methods that use non-local users.
        # these methods are mutually exclusive.
        external_auth_methods = {
            "OAUTH": OAuthHandler,
            "REMOTE_USER": RemoteUserHandler,
        }

        configured_external_auth_methods = [
            k for k in app.config.keys() if k in external_auth_methods.keys()
        ]
        if len(configured_external_auth_methods) > 1:
            raise ValueError(
                "You configured multiple non-local authentication methods: {}. "
                "This is not supported. Please configure only one.".format(
                    configured_external_auth_methods
                )
            )
        elif configured_external_auth_methods:
            auth_method = configured_external_auth_methods[0]
            handler = external_auth_methods[auth_method]

            # set configuration for the auth method to a newly initialized handler
            app.config[auth_method] = handler(app.config[auth_method])

            if app.config.get("ALLOW_ACCOUNT_CREATION"):
                raise ValueError(
                    "When {} is used for authentication, the app "
                    "will not allow account creation".format(auth_method)
                )
            # explicitly set account creation to False
            # used to avoid account conflicts.
            app.config["ALLOW_ACCOUNT_CREATION"] = False

        with app.app_context():
            app.register_blueprint(admin.bp)
            app.register_blueprint(auth.bp)
            app.register_blueprint(team.bp)

    # Ensure the instance folder exists.
    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass

    @app.route("/favicon.ico")
    @app.route("/favicon-<path:path>")
    @app.route("/android-chrome-<path:path>")
    @app.route("/apple-touch-icon.png")
    @app.route("/social-preview.png")
    @app.route("/robots.txt")
    @app.route("/app.webmanifest")
    @app.route("/offline.html")
    @app.route("/service-worker.js")
    def static_from_root(path=None):
        return send_from_directory("build", request.path[1:])

    @app.errorhandler(InternalServerError)
    def error_500(e):
        """Return JSON instead of HTML for HTTP errors."""
        response = e.get_response()
        response.data = json.dumps(
            {
                "code": e.code,
                "name": e.name,
                "description": e.description,
            }
        )
        logging.error(e.description)
        response.content_type = "application/json"
        return response

    @app.route("/signin", methods=["GET"])
    @app.route("/oauth_callback", methods=["GET"])
    @app.route("/reset_password", methods=["GET"])
    @login_remote_user
    def index(**kwargs):
        if isinstance(app.config.get("OAUTH", False), OAuthHandler):
            oauth_params = json.dumps(app.config.get("OAUTH").front_end_params())
        else:
            oauth_params = str(False).lower()

        # Handle terms of agreement config
        terms_config = app.config.get("TERMS_OF_AGREEMENT", False)
        if terms_config is True or (
            isinstance(terms_config, str) and not terms_config.strip()
        ):
            terms_of_agreement = "true"
            terms_text = "I agree to the terms of service"
        elif isinstance(terms_config, str):
            terms_of_agreement = "true"
            terms_text = terms_config
        else:
            terms_of_agreement = "false"
            terms_text = ""

        return render_template(
            "index.html",
            api_url=app.config.get("API_URL", "/"),
            asreview_version=asreview_version,
            authentication=str(app.config.get("AUTHENTICATION", True)).lower(),
            login_info=app.config.get("LOGIN_INFO", ""),
            allow_account_creation=str(
                app.config.get("ALLOW_ACCOUNT_CREATION", True)
            ).lower(),
            email_verification=str(app.config.get("EMAIL_VERIFICATION", False)).lower(),
            oauth=oauth_params,
            post_logout_url=str(app.config.get("POST_LOGOUT_URL", "/signin")).lower(),
            terms_of_agreement=terms_of_agreement,
            terms_text=terms_text,
        )

    @app.route("/", methods=["GET"])
    @app.route("/<path:url>", methods=["GET"])
    @login_remote_user
    def index_protected(**kwargs):
        if app.config.get("AUTHENTICATION", True) and not current_user.is_authenticated:
            return redirect("/signin")

        return index(**kwargs)

    # The task manager needs to be configured if not in testing
    if not (app.testing):
        # I want people to be able to configure the host and port of
        # the task manager by using the env var TASK_MANAGER_ENDPOINT.
        # This var needs to provide host and port in 1 string.
        endpoint = app.config.get("TASK_MANAGER_ENDPOINT", False)
        if endpoint:
            endpoint = endpoint.split(":")
            app.config["TASK_MANAGER_HOST"] = endpoint[0]
            app.config["TASK_MANAGER_PORT"] = int(endpoint[1])

    return app



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Performance metrics for activate learning results."""

__all__ = ["loss", "ndcg"]

import numpy as np


def loss(labels: list[int]):
    """Compute the loss of the labels.

    arguments
    ---------
    labels: list
        List of labels.

    Returns
    -------
    float:
        The loss of the labels.
    """
    Ny = sum(labels)
    Nx = len(labels)
    if Ny == 0 or Nx == Ny:
        raise ValueError("Labels must contain two distinct classes.")
    return float(
        (Ny * (Nx - (Ny - 1) / 2) - np.cumsum(labels).sum()) / (Ny * (Nx - Ny))
    )


def ndcg(labels: list[int]):
    """Compute the Normalized Discounted Cumulative Gain (NDCG)

    Basesd on: https://doi.org/10.48550/arXiv.1304.6480

    Arguments
    ---------
    labels: list
        List of binary labels (0 or 1).

    Returns
    -------
    float:
        The NDCG score.
    """
    Ny = sum(labels)
    Nx = len(labels)
    if Ny == 0 or Nx == Ny:
        raise ValueError("Labels must contain two distinct classes.")

    dcg = np.sum(labels / np.log2(np.arange(2, Nx + 2)))
    idcg = np.sum(1 / np.log2(np.arange(2, Ny + 2)))

    return float(dcg / idcg)



import json
import shutil
import sqlite3
from pathlib import Path

import jsonschema
from sqlalchemy import create_engine

from asreview.data.record import Base
from asreview.database.database import Database
from asreview.project.schema import SCHEMA


def _migrate(project):
    """Migrate an valid project file from version 2 to version 3.

    From version 2 to version 3 the changes to the project file are:

    - ./reviews/<review_id>/results.db is moved to ./results.db
    - ./data_store.db is merged into ./results.db
    - ./reviews/<review_id>/settings_metadata.json is moved to ./settings_meta_data.json
    - ./reviews is deleted.
    In project.json:
    - The `reviews` attribute is moved to `review` and it is turned from a list into a
    single object.
    - The project file version is updated.

    Parameters
    ----------
    Project : str | Path
        Path to the root of the project (unzipped).
    """
    # In project.json, move the `reviews` attribute to `review`.
    config_fp = Path(project, "project.json")
    with open(config_fp) as f:
        project_config = json.load(f)
    project_config["project_file_version"] = 3
    review_config = {}
    reviews = project_config.pop("reviews", [{}])
    old_review_config = reviews[0] if reviews else {}
    review_config["status"] = old_review_config.get("status", "setup")

    # Move the results database to the project root and add the review settings to
    # the project config.
    if old_review_config and "id" in old_review_config:
        review_dir = Path(project, "reviews", old_review_config["id"])
        results_db = Path(review_dir, "results.db")
        if results_db.exists():
            shutil.move(results_db, project)
        review_config_fp = Path(review_dir, "settings_metadata.json")
        if review_config_fp.exists():
            with open(review_config_fp) as f:
                review_config["model"] = json.load(f)
        else:
            review_config["model"] = {}
        tags_config_fp = Path(review_dir, "tags.json")
        if tags_config_fp.exists():
            shutil.copy(tags_config_fp, Path(project, "tags.json"))
            with open(tags_config_fp) as f:
                project_config["tags"] = json.load(f)
    reviews_dir = Path(project, "reviews")
    if reviews_dir.exists():
        shutil.rmtree(reviews_dir)
    project_config["review"] = review_config

    with open(config_fp, "w") as f:
        json.dump(project_config, f)

    _copy_store_to_results(project)
    Path(project, "data_store.db").unlink()

    _update_database_version(project)
    _add_database_triggers(project)


def _copy_store_to_results(project):
    engine = create_engine(f"sqlite:///{str(Path(project, 'results.db'))}")
    Base.metadata.create_all(engine)
    engine.dispose()
    conn = sqlite3.connect(Path(project, "results.db"))
    try:
        cur = conn.cursor()
        data_store_fp = Path(project, "data_store.db")
        cur.execute("ATTACH ? AS data_store", (str(data_store_fp),))
        cur.execute("""
            INSERT INTO record(dataset_row, dataset_id, duplicate_of, title, abstract,
                    authors, keywords, year, doi, url, included, record_id)
            SELECT dataset_row, dataset_id, duplicate_of, title, abstract, authors,
                    keywords, year, doi, url, included, record_id FROM data_store.record
        """)
        conn.commit()
        cur.execute("DETACH DATABASE data_store")
    finally:
        conn.close()


def _update_database_version(project):
    conn = sqlite3.connect(Path(project, "results.db"))
    cur = conn.cursor()
    cur.execute("PRAGMA user_version = 3")
    conn.commit()
    conn.close()


def _add_database_triggers(project):
    with Database(Path(project, "results.db")) as db:
        db._set_results_changes_triggers()


def _validate(project):
    """Validate a migrated v3 project.

    Parameters
    ----------
    project : Path
        Path to the migrated project folder.

    Raises
    ------
    ValueError
        If the migrated project is not valid.
    """
    config_fp = Path(project, "project.json")
    if not config_fp.exists():
        raise ValueError("Migrated project is missing project.json.")

    with open(config_fp) as f:
        config = json.load(f)

    jsonschema.validate(instance=config, schema=SCHEMA)

    if config.get("project_file_version") != 3:
        raise ValueError(
            f"Expected project file version 3, "
            f"got {config.get('project_file_version')}."
        )

    results_db_fp = Path(project, "results.db")
    if not results_db_fp.exists():
        raise ValueError("Migrated project is missing results.db.")

    with Database(results_db_fp) as db:
        db._is_valid()



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["CSVReader"]


import pandas as pd

from asreview.data.base import BaseReader
from asreview.data.record import Record


class CSVReader(BaseReader):
    """CVS file reader."""

    read_format = [".csv", ".tab", ".tsv"]
    write_format = [".csv", ".tsv", ".xlsx"]

    mime_types = {
        "text/csv": [".csv"],
        "text/tab-separated-values": [".tsv", ".tab"],
        "text/plain": [".csv", ".tsv", ".tab"],
    }

    @classmethod
    def read_data(cls, fp):
        """Import dataset.

        Parameters
        ----------
        fp: str, pathlib.Path
            File path to the CSV file.

        Returns
        -------
        list:
            List with entries.
        """
        for encoding in ["utf-8", "ISO-8859-1"]:
            try:
                return pd.read_csv(fp, sep=None, encoding=encoding, engine="python")
            except UnicodeDecodeError:
                # if unicode error, go to next encoding
                continue

        raise UnicodeDecodeError("The encoding of the file is not supported.")


class CSVWriter:
    """CSV file writer."""

    name = "csv"
    label = "CSV (UTF-8)"
    write_format = ".csv"

    @classmethod
    def write_data(cls, df, fp, sep=","):
        """Export dataset.

        Parameters
        ----------
        df: pandas.Dataframe
            Dataframe of all available record data.
        fp: str, NoneType
            Filepath or None for buffer.
        sep: str
            Seperator of the file.

        Returns
        -------
        CSV file
            Dataframe of all available record data.
        """
        return df.to_csv(fp, sep=sep, index=True, date_format="%Y-%m-%d %H:%M:%S")


class ExcelReader(BaseReader):
    """Excel file reader."""

    read_format = [".xlsx"]
    write_format = [".csv", ".tsv", ".xlsx"]

    mime_types = {
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": [".xlsx"],
    }

    @classmethod
    def read_data(cls, fp):
        """Import dataset.

        Parameters
        ----------
        fp: str, pathlib.Path
            File path to the Excel file (.xlsx).

        Returns
        -------
        list:
            List with entries.
        """
        try:
            dfs = pd.read_excel(fp, sheet_name=None)
        except UnicodeDecodeError:
            dfs = pd.read_excel(fp, sheet_name=None, encoding="ISO-8859-1")

        best_sheet = None
        sheet_obj_val = -1
        wanted_columns = Record.get_columns()
        for col_names in cls.__alternative_column_names__.values():
            wanted_columns.extend(col_names)

        for sheet_name in dfs:
            col_names = set([col.lower() for col in list(dfs[sheet_name])])
            obj_val = len(col_names & set(wanted_columns))
            if obj_val > sheet_obj_val:
                sheet_obj_val = obj_val
                best_sheet = sheet_name

        return dfs[best_sheet]


class ExcelWriter:
    """Excel file writer."""

    name = "xlsx"
    label = "Excel"
    write_format = ".xlsx"

    @classmethod
    def write_data(cls, df, fp):
        """Export dataset.

        Parameters
        ----------
        df: pandas.Dataframe
            Dataframe of all available record data.
        fp: str, NoneType
            Filepath or None for buffer.

        Returns
        -------
        Excel file
            Dataframe of all available record data.
        """
        return df.to_excel(fp, index=True, engine="openpyxl")


class TSVWriter:
    """TSV file writer."""

    name = "tsv"
    label = "TSV (UTF-8)"
    write_format = ".tsv"

    @classmethod
    def write_data(cls, df, fp, sep="\t"):
        """Export dataset.

        Parameters
        ----------
        df: pandas.Dataframe
            Dataframe of all available record data.
        fp: str, NoneType
            Filepath or None for buffer.
        sep: str
            Seperator of the file.

        Returns
        -------
        TSV file
            Dataframe of all available record data.
        """
        return df.to_csv(fp, sep=sep, index=True, date_format="%Y-%m-%d %H:%M:%S")



import os

import pytest

import asreview as asr
from asreview.extensions import extensions
from asreview.extensions import load_extension

REQUIRES_AI_MODEL_DEP = ["doc2vec", "embedding-idf", "sbert"]


def test_feature():
    assert len(extensions("models.feature_extractors")) >= 2


@pytest.mark.parametrize("feature_extractor", extensions("models.feature_extractors"))
def test_feature_name(feature_extractor):
    model = load_extension("models.feature_extractors", feature_extractor.name)()
    assert model.name == feature_extractor.name


@pytest.mark.parametrize("feature_extractor", extensions("models.feature_extractors"))
def test_feature_extractor_param(feature_extractor):
    model = load_extension("models.feature_extractors", feature_extractor.name)()
    assert isinstance(model.get_params(), dict)


@pytest.mark.parametrize("feature_extractor", extensions("models.feature_extractors"))
def test_features(tmpdir, feature_extractor):
    data_fp = os.path.join("tests", "demo_data", "generic.csv")

    db = asr.load_dataset(data_fp, dataset_id="test_id")
    model = load_extension("models.feature_extractors", feature_extractor.name)()

    X = model.fit_transform(db.input.get_df())

    assert X.shape[0] == len(db.input)
    assert X.shape[1] > 0



import datetime as dt
from pathlib import Path

from flask import render_template_string
from flask_login import login_user
from flask_mail import Mail
from flask_mail import Message


def has_email_configuration(app):
    return all(
        [
            app.config.get("MAIL_SERVER", False),
            app.config.get("MAIL_USERNAME", False),
            app.config.get("MAIL_PASSWORD", False),
        ]
    )


def get_email_footer_config(app):
    """
    Extract email footer configuration from app config.
    Returns a list of footer paragraphs and processed footer data for templates.
    """

    default_footer = {"paragraph_1": "Regards,", "paragraph_2": "The ASReview Team"}

    # Get email_footer section from config
    email_footer_config = app.config.get("EMAIL_FOOTER", default_footer)

    # if no paragraphs are found
    if len(email_footer_config.keys()) == 0:
        email_footer_config = default_footer

    # Extract paragraphs from config (paragraph_1, paragraph_2, etc.)
    paragraphs = []
    for key in sorted(email_footer_config.keys()):
        if key.startswith("paragraph_"):
            paragraphs.append(email_footer_config[key])

    # Create HTML footer
    html_footer = "\n\n".join([f"<p>{paragraph}</p>" for paragraph in paragraphs])

    # Create text footer
    txt_footer = "\n\n".join(paragraphs)

    return {"html_footer": html_footer, "txt_footer": txt_footer}


def perform_login_user(user, app):
    """Helper function to login a user"""
    return login_user(
        user,
        remember=True,
        duration=dt.timedelta(days=app.config.get("LOGIN_DURATION", 31)),
    )


def send_forgot_password_email(user, cur_app):
    # do not send email in test environment
    if not cur_app.testing:
        # set name
        name = user.name or "ASReview user"

        # get footer configuration
        footer_config = get_email_footer_config(cur_app)

        # create a mailer
        mailer = Mail(cur_app)
        # open templates as string and render
        root_path = Path(cur_app.root_path)
        with open(root_path / "templates" / "emails" / "forgot_password.html") as f:
            html_text = render_template_string(
                f.read(),
                name=name,
                token=user.token,
                footer_html=footer_config["html_footer"],
            )
        with open(root_path / "templates" / "emails" / "forgot_password.txt") as f:
            txt_text = render_template_string(
                f.read(),
                name=name,
                token=user.token,
                footer_txt=footer_config["txt_footer"],
            )
        # create message
        msg = Message("ASReview: forgot password", recipients=[user.email])
        msg.body = txt_text
        msg.html = html_text
        return mailer.send(msg)


def send_confirm_account_email(user, cur_app, email_type="create"):
    # do not send email in test environment
    if not cur_app.testing:
        # get necessary information out of user object
        name = user.name or "ASReview user"

        # get footer configuration
        footer_config = get_email_footer_config(cur_app)

        # create a mailer
        mailer = Mail(cur_app)
        # open templates as string and render
        root_path = Path(cur_app.root_path)
        with open(root_path / "templates" / "emails" / "confirm_account.html") as f:
            html_text = render_template_string(
                f.read(),
                name=name,
                token=user.token,
                email_type=email_type,
                footer_html=footer_config["html_footer"],
            )
        with open(root_path / "templates" / "emails" / "confirm_account.txt") as f:
            txt_text = render_template_string(
                f.read(),
                name=name,
                token=user.token,
                email_type=email_type,
                footer_txt=footer_config["txt_footer"],
            )
        # create message
        msg = Message("ASReview: please confirm your account", recipients=[user.email])
        msg.body = txt_text
        msg.html = html_text
        return mailer.send(msg)



import numpy as np
import pandas as pd
from sqlalchemy import NullPool
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from asreview.project.migration.legacy_v2.record import Base
from asreview.project.migration.legacy_v2.record import Record

CURRENT_DATASTORE_VERSION = 0


class DataStore:
    """Data store to hold user input data.

    Data input always happends via the record class. This means that if you want to add
    data to the data store, you will first need to clean it, make sure it has the
    correct columns and make sure it passes the validations defined in the record class.

    Getting data from the store can happen in rows or in columns. If you read rows, you
    will get record objects as response. If you read columns, you will get pandas
    objects. If you ask for a single column you get a pandas Series, and if you ask for
    multiple columns you get a pandas DataFrame.

    DataStore uses an SQLite database in the backend and SQLAlchemy ORM to interact with
    the database."""

    def __init__(self, fp, record_cls=Record):
        """Initialize the data store.

        Parameters
        ----------
        fp : str | Path
            Location of the database file. If `fp == ":memory:"`, the data store will be
            in memory.
        record_cls : asreview.data.record.Base, optional
            The record class to use. The record class specifies which fields each record
            can have, field validation and more properties of the database. See
            `asreview.data.record`. By default uses `asreview.data.record.Record`.
        """
        self.fp = fp
        # If the sqlite database is in memory, we should use the default
        # SingleThreadPool poolclass, because any poolclass with multiple threads will
        # have different instances of the database for each thread.
        if fp == ":memory:":
            poolclass = None
        else:
            poolclass = NullPool
        # I'm using NullPool here, indicating that the engine should use a connection
        # pool, but just create and dispose of a connection every time a request comes.
        # This makes it very easy dispose of the engine, but is less efficient.
        # I was getting errors when running tests that try to clean up behind them,
        # and this solves those errors. We can change this back to a connection pool at
        # some later moment by properly looking at how to close everything.
        self.engine = create_engine(f"sqlite:///{self.fp}", poolclass=poolclass)
        # I put expire_on_commit=False, so that after you put records in the database,
        # you can still use them in your code without having access to the database.
        # The downside is that if you use the record after committing it to the database
        # and another mutation happens to the database, your record might be out of
        # date. See https://docs.sqlalchemy.org/en/20/orm/session_api.html#sqlalchemy.orm.Session.params.expire_on_commit
        self.Session = sessionmaker(self.engine, expire_on_commit=False)
        self.record_cls = record_cls
        self._columns = self.record_cls.get_columns()
        self._pandas_dtype_mapping = self.record_cls.get_pandas_dtype_mapping()

    @property
    def columns(self):
        return self._columns

    @property
    def pandas_dtype_mapping(self):
        """Mapping {column name: pandas data type}"""
        return self._pandas_dtype_mapping

    def create_tables(self):
        """Initialize the tables containing the data.

        If you are creating a new data store, you will need to call this method before
        adding data to the data store."""
        Base.metadata.create_all(self.engine)

    def add_records(self, records):
        """Add records to the data store.

        Parameters
        ----------
        records : list[self.record_cls]
            List of records to add to the store.
        """
        # SQLite makes an autoincremented primary key column start at 1. We want it to
        # start at 0, so that the record_id is equal to the row number of the record in
        # feature matrix. By making sure that the first record has record_id 0, we force
        # the autoincremented column to start at 0.
        if self.is_empty():
            records[0].record_id = 0

        with self.Session() as session, session.begin():
            session.add_all(records)

    def delete_record(self, record_id):
        with self.Session() as session, session.begin():
            record = session.get(self.record_cls, record_id)
            if record is None:
                raise ValueError(
                    f"DataStore does not contain a record with record_id {record_id}"
                )
            session.delete(record)

    def __len__(self):
        with self.Session() as session:
            return session.query(self.record_cls).count()

    def __getitem__(self, item):
        # We allow a string or a list of strings as input. If the input is a string we
        # return that column as a pandas series. If the input is a list of strings we
        # return a pandas DataFrame containing those columns. This way the output you
        # get is the same if you do __getitem__ on a DataStore instance or on a pandas
        # DataFrame containing the same data.
        if isinstance(item, str):
            columns = [item]
        else:
            columns = item
        with self.engine.connect() as con:
            df = pd.read_sql(
                self.record_cls.__tablename__,
                con,
                columns=columns,
                dtype=self.pandas_dtype_mapping,
            )
        if isinstance(item, str):
            return df[item]
        else:
            return df

    def __contains__(self, item):
        return item in self.columns

    def is_empty(self):
        with self.Session() as session:
            return session.query(self.record_cls).first() is None

    def get_records(self, record_id):
        """Get the records with the given record identifiers.

        Parameters
        ----------
        record_id : int | list[int]
            Record identifier or list record identifiers.

        Returns
        -------
        asreview.data.record.Record | list[asreview.data.record.Record] | None
        """
        if isinstance(record_id, np.integer):
            record_id = record_id.item()

        with self.Session() as session:
            if isinstance(record_id, int):
                return (
                    session.query(self.record_cls)
                    .filter(self.record_cls.record_id == record_id)
                    .first()
                )
            else:
                records = (
                    session.query(self.record_cls)
                    .filter(self.record_cls.record_id.in_(record_id))
                    .all()
                )

                record_id_to_position = {id: i for i, id in enumerate(record_id)}
                return sorted(records, key=lambda r: record_id_to_position[r.record_id])

    def get_df(self):
        """Get all data from the data store as a pandas DataFrmae.

        Returns
        -------
        pd.DataFrame
        """
        with self.engine.connect() as con:
            return pd.read_sql(
                self.record_cls.__tablename__,
                con,
                dtype=self.pandas_dtype_mapping,
            )



import pytest
from jinja2.exceptions import TemplateNotFound

from asreview.webapp._authentication.models import User
from asreview.webapp.tests.utils.misc import custom_remote_auth_headers

# ###################
# LOGIN FAILS
# ###################


@pytest.mark.parametrize("uri", ["/", "/signup"])
def test_no_login_without_header(client_remote_auth, uri):
    custom_headers = custom_remote_auth_headers(
        remote_user=""
    )  # setup the REMOTE_AUTH_SECRET header
    r = client_remote_auth.get(uri, **custom_headers)
    assert r.status_code == 302
    assert r.location == "/signin"


def test_no_login_without_secret(client_remote_auth, uri="/"):
    custom_headers = custom_remote_auth_headers(remote_user="foo", secret=None)

    response = client_remote_auth.get(uri, follow_redirects=True, **custom_headers)
    assert "REMOTE_AUTH_SECRET did not match" in response.get_json()["message"]
    assert response.status_code == 401


# ###################
# LOGIN SUCCESSFUL
# ###################


@pytest.mark.parametrize("uri", ["/", "/signup"])
def test_login_with_header(client_remote_auth, uri):
    user_identifier = "foo@dev.bar"
    custom_headers = custom_remote_auth_headers(remote_user=user_identifier)

    def get_uri(path):
        return client_remote_auth.get(uri, follow_redirects=True, **custom_headers)

    def find_user(identifier):
        return User.query.filter(User.identifier == identifier).one_or_none()

    user = find_user(user_identifier)
    assert user is None

    try:
        get_uri(uri)
    except TemplateNotFound:
        pass

    user = find_user(user_identifier)
    assert user
    assert user.is_authenticated
    assert user.email == "foo@dev.bar"
    assert user.affiliation == "UU"


# ###################
# API
# ###################


def test_api_returns_401_without_header(client_remote_auth):
    r = client_remote_auth.get("/auth/user")
    assert r.status_code == 401


def test_api_returns_user_with_header(client_remote_auth):
    custom_headers = custom_remote_auth_headers(remote_user="foo@dev.bar")

    r = client_remote_auth.get("/auth/user", **custom_headers)
    user_info = r.get_json()
    assert user_info["id"] == 1
    assert user_info["name"] == "foo"



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sklearn.base import BaseEstimator
from sklearn.utils.class_weight import compute_sample_weight as _compute_sample_weight

__all__ = [
    "Balanced",
]


class Balanced(BaseEstimator):
    """Balanced sample weight

    Parameters
    ----------
    ratio: float
        The ratio of the number of samples in class 0 to the number of samples
        in class 1.

    """

    name = "balanced"
    label = "Balanced Sample Weight"

    def __init__(self, ratio=1.0):
        self.ratio = ratio

    def compute_sample_weight(self, y):
        if len(set(y)) != 2:
            raise ValueError("Only binary classification is supported.")

        weights = _compute_sample_weight(
            {1: 1.0, 0: sum(y == 1) / (self.ratio * sum(y == 0))}, y=y
        )
        return weights * (len(y) / sum(weights))



from pathlib import Path

import pandas as pd
import pytest

from asreview.data.loader import _get_reader
from asreview.data.loader import _get_writer


@pytest.mark.parametrize(
    "test_file,columns",
    [
        ("ris_issue_992.txt", ["title"]),
        ("ris_issue_1099.txt", ["primary_title"]),
        ("baseline_tag-notes.ris", ["title", "notes"]),
        ("baseline_tag-notes_labels.ris", ["title", "included"]),
    ],
)
def test_asreview_ris(test_file, columns, tmpdir):
    fp_in = Path("tests", "demo_data", test_file)
    data = _get_reader(fp_in).read_data(fp_in)
    if "included" in data:
        data["asreview_label"] = data["included"]

    tmp_ris_fp_out = Path(tmpdir, "tmp.ris")
    _get_writer(tmp_ris_fp_out).write_data(data, tmp_ris_fp_out)
    written_data = _get_reader(tmp_ris_fp_out).read_data(tmp_ris_fp_out)

    for col in columns:
        pd.testing.assert_series_equal(data[col], written_data[col])
